@@ -25,7 +25,6 @@ pub enum PaymentState<ProcessArgs> {
     Idle,
     Processing(ProcessArgs),
     Succeeded,
-    Failed,
     Cancelled,
 }
 
@@ -60,8 +59,8 @@ pub trait PaymentProcessor {
         args: &Self::ProcessArgs,
     ) -> Result<ProcessingResult<Self::CorrelationID, Self::PrivateData, Self::PublicData>, Error>;
 
-    // this is called synchronously and only once per cancel_payment request
-    async fn cancel_payment(
+    // this is called synchronously and only once per revert request
+    async fn revert(
         &mut self,
         payment: &Payment<
             Self::CorrelationID,
@@ -231,13 +230,15 @@ pub async fn queue_for_processing<C: Context + Send>(
             "PAYMENT_PROCESSING",
             "processing is in progress",
         )),
-        PaymentState::Failed | PaymentState::Succeeded | PaymentState::Cancelled => Err(
-            new_application_error("PAYMENT_COMPLETED", "the payment has already completed"),
-        ),
+        PaymentState::Succeeded | PaymentState::Cancelled => Err(new_application_error(
+            "PAYMENT_COMPLETED",
+            "the payment has already completed",
+        )),
     }
 }
 
 pub async fn process<C: Context + Send>(
+    internal_ctx: &InternalContext,
     ctx: &mut C,
     payment: &mut Payment<
         <C as PaymentProcessor>::CorrelationID,
@@ -252,18 +253,32 @@ pub async fn process<C: Context + Send>(
             let result = ctx.process_payment(payment, args).await?;
             payment.data = result.data;
             match result.state {
-                InternalPaymentState::Incomplete => payment.state = PaymentState::Idle,
+                InternalPaymentState::Incomplete | InternalPaymentState::Failed => {
+                    payment.state = PaymentState::Idle
+                }
                 InternalPaymentState::Succeeded => payment.state = PaymentState::Succeeded,
                 InternalPaymentState::Cancelled => payment.state = PaymentState::Cancelled,
-                InternalPaymentState::Failed => payment.state = PaymentState::Failed,
             }
 
             ctx.update_payment(payment).await?;
 
             match payment.state {
-                PaymentState::Succeeded | PaymentState::Cancelled | PaymentState::Failed => {
-                    checkout.complete_payment(ctx).await
+                PaymentState::Succeeded => {
+                    internal_ctx
+                        .payment_processing_topic
+                        .clone()
+                        .publish(&payment.checkout_id)
+                        .await
+                        .map_err(|_| {
+                            new_application_error(
+                                "UPSTREAM_ERROR",
+                                "couldn't send message to payment processing topic",
+                            )
+                        })?;
+
+                    checkout.complete(ctx).await
                 }
+                PaymentState::Cancelled => checkout.revert(ctx).await,
                 _ => Ok(()),
             }
         }
@@ -279,11 +294,12 @@ pub async fn cancel<C: Context + Send>(
         <C as PaymentProcessor>::PublicData,
         <C as PaymentProcessor>::ProcessArgs,
     >,
+    checkout: &mut Checkout,
 ) -> Result<(), Error> {
     match &payment.state {
         PaymentState::Cancelled => Ok(()),
         PaymentState::Idle => {
-            let cancellation_result = ctx.cancel_payment(payment).await?;
+            let cancellation_result = ctx.revert(payment).await?;
             if !cancellation_result.success {
                 return Err(new_application_error(
                     "CANCELLATION_UNAVAILABLE",
@@ -292,13 +308,15 @@ pub async fn cancel<C: Context + Send>(
             }
 
             payment.state = PaymentState::Cancelled;
-            ctx.update_payment(payment).await
+            ctx.update_payment(payment).await?;
+
+            checkout.revert(ctx).await
         }
         PaymentState::Processing(_) => Err(new_application_error(
             "PAYMENT_PROCESSING",
             "the payment is processing",
         )),
-        PaymentState::Failed | PaymentState::Succeeded => Err(new_application_error(
+        PaymentState::Succeeded => Err(new_application_error(
             "PAYMENT_COMPLETED",
             "the payment has already completed",
         )),
